/* outchannels.cpp - MuSE
   (c)2000 Denis Roio aka jaromil
   see COPYING for licensing stuff

   different classes for different OUT channels
   they are instantiated and used by the Stream_mixer class (jmixer.cpp)
*/

#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>

#include "generic.h"
#include "outchannels.h"
#include "jutils.h"


/* I think a tighter bound could be:  (mt, March 2000)
 * MPEG1:
 *    num_samples*(bitrate/8)/samplerate + 4*1152*(bitrate/8)/samplerate + 512
 * MPEG2:
 *    num_samples*(bitrate/8)/samplerate + 4*576*(bitrate/8)/samplerate + 256
 */

#define ENCBUFFER_SIZE 32768

static unsigned char buffer[ENCBUFFER_SIZE];
static unsigned long int encoded;

/* mp3 encoder */

LameEnc::LameEnc() {
  initialized = false;
}

/* note: num is number of samples in the L (or R) channel
   not the total number of samples in pcm[] */
void LameEnc::encode(short int *pcm, int num) {
  encoded = lame_encode_buffer_interleaved(enc_flags, pcm, num,
					   (unsigned char *)buffer, ENCBUFFER_SIZE);
}

void LameEnc::flush() {
  lame_encode_flush(enc_flags,(unsigned char *)buffer,ENCBUFFER_SIZE);
}

void LameEnc::init(int nbps, int freq, int mode, int quality, int lowpass, int highpass) {
  enc_flags = lame_init();
  lame_set_num_samples(enc_flags,1024); /* we get 1024 samples */
  lame_set_num_channels(enc_flags,2); /* the mixed input stream is stereo */
  lame_set_in_samplerate(enc_flags,SAMPLE_RATE); /* the mixed input stream is 44khz */
  if(freq>0) lame_set_out_samplerate(enc_flags,freq);
  bps = nbps;
  lame_set_brate(enc_flags,bps);
  lame_set_mode(enc_flags,(enum MPEG_mode_e)mode); /* 0,1,2,3 stereo,jstereo,dual channel,mono */
  lame_set_quality(enc_flags,quality); /* 0..9 0=best 9=worst (suggested: 2, 5, 7) */
  
  lame_set_error_protection(enc_flags,1); /* 2 bytes per frame for a CRC checksum */
  
  /* lame chooses for us frequency filtering when values are 0 */
  lame_set_lowpassfreq(enc_flags,lowpass);
  lame_set_highpassfreq(enc_flags,highpass);

  lame_init_params(enc_flags);
  initialized = true;
}

char *LameEnc::version() {
  ver = (char *)get_lame_version();
  return(ver);
}

void LameEnc::close() {
  lame_close(enc_flags);
}

LameEnc::~LameEnc() {
  if(initialized) close();
}

/* IcecastOut */

IcecastOut::IcecastOut() {
  initialized = false;
  running = false;
  pass = NULL;
  mountpoint = NULL;
  name = NULL;
  url = NULL;
  description = NULL;
}

IcecastOut::~IcecastOut() {
  if(running)
    stop();
  if(pass!=NULL) free(pass);
  if(mountpoint!=NULL) free(mountpoint);
  if(name!=NULL) free(name);
  if(url!=NULL) free(url);
  if(description!=NULL) free(description);
}

bool IcecastOut::start() {
  int res;

  if(initialized&&!running) {
    shout_init_connection(&ice_conn);
    ice_conn.ip = ip;
    ice_conn.port = port;
    ice_conn.password = pass;
    ice_conn.mount = mountpoint;
    ice_conn.bitrate = bps*1000;
    ice_conn.name = name;
    ice_conn.url = url;
    ice_conn.description = description;
    ice_conn.icy_compat = login;

    notice("Contacting icecast server %s:%u",ip,port);
    
    res = shout_connect(&ice_conn);
    if(res) { 
      act("allright, we are streaming!");
      running = true;
    } else {
      shout_strerror(&ice_conn,ice_conn.error,errormsg,ERRORMSG);
      warning("%s",errormsg);
      running = false;
    }

    return(running);
  } else
    return(false); /* icecast connection not initialized or already running */
}

void IcecastOut::stop() {
  if(running) {
    shout_disconnect(&ice_conn);
    running = false;
  }
}

bool IcecastOut::set_server(char *naddress, int nport, char *npass, char *nmountpoint, int nbps, int nlogin) {
  struct sockaddr_in server_addr;
  struct hostent *h;

  stop();

  memset(&server_addr, 0, sizeof(struct sockaddr_in));
  server_addr.sin_family = AF_INET;

#ifdef HAVE_INET_ATON
  inet_aton(naddress,&server_addr.sin_addr);
#else
  server_addr.sin_addr.s_addr=inet_addr(naddress);
#endif

  if(!(h = gethostbyname(naddress)))
     return(false);
     
  server_addr.sin_family = AF_INET;
  memcpy(&server_addr.sin_addr, h->h_addr, h->h_length);    
  
  sprintf(ip,"%s",
	  inet_ntoa(*(struct in_addr *) &server_addr.sin_addr.s_addr));
  
  port = nport;

  if(pass!=NULL) free(pass);
  pass = strdup(npass);
  if(mountpoint!=NULL) free(mountpoint);
  mountpoint = strdup(nmountpoint);
  bps = nbps;
  login = nlogin;
  
  initialized = true;
  return(initialized);
}

void IcecastOut::set_meta(char *nname, char *nurl, char *ndesc) {
  if(name!=NULL) free(name);
  name = strdup(nname);
  
  if(url!=NULL) free(nurl);
  url = strdup(nurl);

  if(description!=NULL) free(ndesc);
  description = strdup(ndesc);
}

int IcecastOut::send() {
  if(!running) return(0);
  if(encoded<=0) return(0);

  int res;
  sync();
  res = shout_send_data(&ice_conn, buffer, encoded);
  if(!res) {
    shout_strerror(&ice_conn,ice_conn.error,errormsg,ERRORMSG);
    warning("%s",errormsg);
    return(-1);
  }

  return(res);
}

void IcecastOut::sync() {
  shout_sleep(&ice_conn);
}

/* Mp3fileOut */

Mp3fileOut::Mp3fileOut() { 
  opened = false;
  filename = NULL;
}

bool Mp3fileOut::openmp3(char *file) {

  if(opened)
    closemp3();
  
  if(filename!=NULL) free(filename);
  filename = strdup(file);
  
  /* creates a file write only non blocking (-rw-rw-r--) */
  fd = fopen(filename,"w");

  if(fd!=NULL)
    opened = true;

  return(opened);
}

int Mp3fileOut::writemp3() {
  int res = -1;
  
  if(opened&&(encoded>0)) {
    syncmp3();
    res = fwrite(buffer,1,encoded,fd);
  }
  return(res);
}

void Mp3fileOut::closemp3() {
  if(opened) {
    fclose(fd);
    opened = false;
  }
}

bool Mp3fileOut::syncmp3() {
  if(fflush(fd)!=0)
    return(false);
  else
    return(true);
}

Mp3fileOut::~Mp3fileOut() {
  if(opened)
    closemp3();
  if(filename!=NULL) free(filename);
}
